<div class='alert alert-{% if passed %}success{% else %}error{% endif %}'>
  {% if passed %}
  <svg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 24 24'><title>check</title>
    <path d='M21,7L9,19L3.5,13.5L4.91,12.09L9,16.17L19.59,5.59L21,7Z' />
  </svg>
  <strong>Test suite passed.</strong>

  {% else %}
  <svg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 24 24'><title>close</title>
    <path
      d='M19,6.41L17.59,5L12,10.59L6.41,5L5,6.41L10.59,12L5,17.59L6.41,19L12,13.41L17.59,19L19,17.59L13.41,12L19,6.41Z' />
  </svg>
  <strong>Test suite failed.</strong>
  <span>To debug your failing test and diagnose the issue, please run the Giskard server (<a
    href='https://docs.giskard.ai/en/latest/guides/installation_app/index.html'>see documentation</a>)</span>
  {% endif %}
</div>
