<div class='card-container'>
  {% for test_name, result, params in test_results %}
  <div
    class='test-card test-card-{% if result.passed %}passed{% elif result.is_error %}error{% else %}failed{% endif %}'>
    <div class='test-card-row'>
      <span class='test-name'>Test {{ test_name }}</span>
      <div class='spacer'></div>
      {% if result.metric is not none %}
      <span class='metric'>Measured <strong>Metric = {{result.metric|round(5)}}</strong></span>
      {% endif %}
      <span class='test-result'>
      {% if result.passed %}
        <svg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 24 24'><title>check</title>
          <path d='M21,7L9,19L3.5,13.5L4.91,12.09L9,16.17L19.59,5.59L21,7Z' />
        </svg>
        Passed
      {% elif result.is_error %}
        <svg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 24 24'><title>alert-outline</title>
          <path d='M12,2L1,21H23M12,6L19.53,19H4.47M11,10V14H13V10M11,16V18H13V16' />
        </svg>
        Error
      {% else %}
        <svg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 24 24'><title>close</title>
          <path
            d='M19,6.41L17.59,5L12,10.59L6.41,5L5,6.41L10.59,12L5,17.59L6.41,19L12,13.41L17.59,19L19,17.59L13.41,12L19,6.41Z' />
        </svg>
        Failed
      {% endif %}
    </span>
    </div>
    <div class='test-card-row'>
      {% for param_name, param_value in params.items() %}
      <div class='param'>
        <span>{{param_name}}</span>
        <span class='param-value'>{{param_value}}</span>
      </div>
      {% endfor %}
    </div>
  </div>
  {% endfor %}
</div>
